@@ -21,16 +21,24 @@ DEFAULT_ALIGNMENT = 1024
 PAGE_ALIGNMENT = 64
 
 import argparse
+import os
+
+NOCRC=False
+if "NOCRC" in os.environ and len(os.environ["NOCRC"]) != 0:
+	NOCRC=os.environ["NOCRC"]
+
 
 def get_argparser():
 	parser = argparse.ArgumentParser(description="Load an app onto the device from a hex file.")
 	parser.add_argument("--targetId", help="The device's target ID (default is Ledger Blue)", type=auto_int)
+	parser.add_argument("--targetVersion", help="Set the chip target version")
 	parser.add_argument("--fileName", help="The application hex file to be loaded onto the device")
 	parser.add_argument("--icon", help="The icon content to use (hex encoded)")
 	parser.add_argument("--curve", help="""A curve on which BIP 32 derivation is locked ("secp256k1", "prime256r1", or
 "ed25519"), can be repeated""", action='append')
 	parser.add_argument("--path", help="""A BIP 32 path to which derivation is locked (format decimal a'/b'/c), can be
 repeated""", action='append')
+	parser.add_argument("--path_slip21", help="""A SLIP 21 path to which derivation is locked""", action='append')
 	parser.add_argument("--appName", help="The name to give the application after loading it")
 	parser.add_argument("--signature", help="A signature of the application (hex encoded)")
 	parser.add_argument("--signApp", help="Sign application with provided signPrivateKey", action='store_true')
@@ -47,10 +55,13 @@ a random one will be generated)""")
 	parser.add_argument("--tlv", help="Use install parameters for all variable length parameters", action='store_true')
 	parser.add_argument("--dataSize", help="The code section's size in the provided hex file (to separate data from code, if not provided the whole allocated NVRAM section for the application will remain readonly.", type=auto_int)
 	parser.add_argument("--appVersion", help="The application version (as a string)")
-	parser.add_argument("--offline", help="Request to only output application load APDUs", action="store_true")
+	parser.add_argument("--offline", help="Request to only output application load APDUs into given filename")
+	parser.add_argument("--offlineText", help="Request to only output application load APDUs into given filename in text mode", action='store_true')
 	parser.add_argument("--installparamsSize", help="The loaded install parameters section size (when parameters are already included within the .hex file.", type=auto_int)
 	parser.add_argument("--tlvraw", help="Add a custom install param with the hextag:hexvalue encoding", action='append')
 	parser.add_argument("--dep", help="Add a dependency over an appname[:appversion]", action='append')
+	parser.add_argument("--nocrc", help="Skip CRC generation when loading", action='store_true')
+
 	return parser
 
 def auto_int(x):
@@ -72,6 +83,12 @@ def parse_bip32_path(path, apilevel):
 						result = result + struct.pack(">I", 0x80000000 | int(element[0]))
 		return result
 
+def parse_slip21_path(path):
+	import struct
+	result = struct.pack('>B', 0x80 | (len(path) + 1))
+	result = result + b'\x00' + string_to_bytes(path)
+	return result
+
 def string_to_bytes(x):
 	import sys
 	if sys.version_info.major == 3:
@@ -94,13 +111,15 @@ if __name__ == '__main__':
 	args = get_argparser().parse_args()
 
 	if args.apilevel == None:
-		args.apilevel = 5
+		args.apilevel = 10
 	if args.targetId == None:
 		args.targetId = 0x31000002
 	if args.fileName == None:
 		raise Exception("Missing fileName")
 	if args.appName == None:
 		raise Exception("Missing appName")
+	if args.path_slip21 != None and args.apilevel < 10:
+		raise Exception("SLIP 21 path not supported using this API level")
 	if args.appFlags == None:
 		args.appFlags = 0
 	if args.rootPrivateKey == None:
@@ -129,12 +148,20 @@ if __name__ == '__main__':
 			else:
 				raise Exception("Unknown curve " + curve)
 
-	if args.apilevel >= 5:
+	if args.apilevel >= 5:		
+		if (args.path_slip21 != None):
+			curveMask |= 0x08
 		path += struct.pack('>B',curveMask)
 		if args.path != None:
 			for item in args.path:
 				if len(item) != 0:
 					path += parse_bip32_path(item, args.apilevel)
+		if args.path_slip21 != None:
+			for item in args.path_slip21:
+				if len(item) != 0:
+					path += parse_slip21_path(item)
+			if (args.path == None) or ((len(args.path) == 1) and (len(args.path[0]) == 0)):
+				path += struct.pack('>B', 0) # Unrestricted, authorize all paths for regular derivation
 	else:
 		if args.curve != None:
 			print("Curve not supported using this API level, ignoring")
@@ -173,6 +200,24 @@ if __name__ == '__main__':
 			secret = getDeployedSecretV1(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId)
 		else:
 			secret = getDeployedSecretV2(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId)
+	else:
+		fileTarget = open(args.offline, "wb")
+		class FileCard():
+			def __init__(self, target):
+				self.target = target
+			def exchange(self, apdu):
+				if (args.apdu):
+					print(binascii.hexlify(apdu))
+				apdu = binascii.hexlify(apdu)
+				if sys.version_info.major == 2:
+					self.target.write(str(apdu) + '\n')
+				else:
+					self.target.write(apdu + '\n'.encode())
+				return bytearray([])
+			def apduMaxDataSize(self):
+				# ensure to allow for encryption of those apdu afterward
+				return 240
+		dongle = FileCard(fileTarget)
 
 	loader = HexLoader(dongle, 0xe0, not(args.offline), secret, cleardata_block_len=cleardata_block_len)
 
@@ -251,7 +296,7 @@ if __name__ == '__main__':
 		loader.createAppNoInstallParams(args.appFlags, appLength, args.appName, args.icon, path, None, None, string_to_bytes(args.appVersion))
 
 
-	hash = loader.load(0x0, 0xF0, printer)
+	hash = loader.load(0x0, 0xF0, printer, targetId=args.targetId, targetVersion=args.targetVersion, doCRC=not (args.nocrc or NOCRC)) 
 
 	print("Application full hash : " + hash)
 
